{
***************************************************
* A binary compatible GOST implementation         *
* written by Dave Barton (davebarton@bigfoot.com) *
***************************************************
* 64bit block encryption                          *
* 256bit key size                                 *
***************************************************
}
unit GOST;

interface
uses
  Sysutils, Tools;

type
  TGOSTData= record
    InitBlock: array[0..7] of byte;    { initial IV }
    LastBlock: array[0..7] of byte;    { current IV }
    XKey: array[0..7] of DWord;
  end;

function GOSTSelfTest: boolean;
  { performs a self test on this implementation }
procedure GOSTInit(var Data: TGOSTData; Key: pointer; Len: integer; IV: pointer);
  { initializes the TGOSTData structure with the key information and IV if applicable }
procedure GOSTBurn(var Data: TGOSTData);
  { erases all information about the key }

procedure GOSTEncryptECB(const Data: TGOSTData; InData, OutData: pointer);
  { encrypts the data in a 64bit block using the ECB mode }
procedure GOSTEncryptCBC(var Data: TGOSTData; InData, OutData: pointer);
  { encrypts the data in a 64bit block using the CBC chaining mode }
procedure GOSTEncryptOFB(var Data: TGOSTData; InData, OutData: pointer);
  { encrypts the data in a 64bit block using the OFB chaining mode }
procedure GOSTEncryptCFB(var Data: TGOSTData; InData, OutData: pointer; Len: integer);
  { encrypts Len bytes of data using the CFB chaining mode }
procedure GOSTEncryptOFBC(var Data: TGOSTData; InData, OutData: pointer; Len: integer);
  { encrypts Len bytes of data using the OFB counter chaining mode }

procedure GOSTDecryptECB(const Data: TGOSTData; InData, OutData: pointer);
  { decrypts the data in a 64bit block using the ECB mode }
procedure GOSTDecryptCBC(var Data: TGOSTData; InData, OutData: pointer);
  { decrypts the data in a 64bit block using the CBC chaining mode }
procedure GOSTDecryptOFB(var Data: TGOSTData; InData, OutData: pointer);
  { decrypts the data in a 64bit block using the OFB chaining mode }
procedure GOSTDecryptCFB(var Data: TGOSTData; InData, OutData: pointer; Len: integer);
  { decrypts Len bytes of data using the CFB chaining mode }
procedure GOSTDecryptOFBC(var Data: TGOSTData; InData, OutData: pointer; Len: integer);
  { decrypts Len bytes of data using the OFB counter chaining mode }

procedure GOSTReset(var Data: TGOSTData);
  { resets the chaining mode information }

{******************************************************************************}
implementation
{$R-}

{$I Gost.Inc}

function GOSTSelfTest;
const
  Key: array[0..31] of byte=
    ($BE,$5E,$C2,$00,$6C,$FF,$9D,$CF,$52,$35,$49,$59,$F1,$FF,$0C,$BF,$E9,$50,$61,$B5,$A6,$48,$C1,$03,$87,$06,$9C,$25,$99,$7C,$06,$72);
  InBlock: array[0..7] of byte=
    ($0D,$F8,$28,$02,$B7,$41,$A2,$92);
  OutBlock: array[0..7] of byte=
    ($07,$F9,$02,$7D,$F7,$F7,$DF,$89);
var
  Block: array[0..7] of byte;
  Data: TGOSTData;
begin
  GOSTInit(Data,@Key,Sizeof(Key),nil);
  GOSTEncryptECB(Data,@InBlock,@Block);
  Result:= CompareMem(@Block,@OutBlock,Sizeof(Block));
  GOSTDecryptECB(Data,@Block,@Block);
  Result:= Result and CompareMem(@Block,@InBlock,Sizeof(Block));
  GOSTBurn(Data);
end;

procedure GOSTInit;
var
  i: integer;
  userkey: PByteArray;
begin
  if (Len<> 32) then
    raise Exception.Create('GOST: Invalid key length');
  UserKey:= Key;
  with Data do
  begin
    if IV= nil then
    begin
      FillChar(InitBlock,8,0);
      FillChar(LastBlock,8,0);
    end
    else
    begin
      Move(IV^,InitBlock,8);
      Move(IV^,LastBlock,8);
    end;
    for i:= 0 to 7 do
      XKey[i]:= (UserKey[4*i+3] shl 24) or (UserKey[4*i+2] shl 16) or
        (UserKey[4*i+1] shl 8) or (UserKey[4*i+0]);
  end;
end;

procedure GOSTBurn;
begin
  FillChar(Data,Sizeof(Data),0);
end;

function f(const x: DWord): DWord;
begin
  Result:= sTable[3,x shr 24] xor sTable[2,(x shr 16) and $FF] xor sTable[1,(x shr 8) and $FF] xor sTable[0,x and $FF];
end;

procedure GOSTEncryptECB;
var
  n1, n2: DWord;
  i: integer;
begin
  Move(InData^,n1,4);
  Move(pointer(integer(InData)+4)^,n2,4);
  for i:= 0 to 2 do
  begin
    n2:= n2 xor f(n1+Data.XKey[0]);
    n1:= n1 xor f(n2+Data.XKey[1]);
    n2:= n2 xor f(n1+Data.XKey[2]);
    n1:= n1 xor f(n2+Data.XKey[3]);
    n2:= n2 xor f(n1+Data.XKey[4]);
    n1:= n1 xor f(n2+Data.XKey[5]);
    n2:= n2 xor f(n1+Data.XKey[6]);
    n1:= n1 xor f(n2+Data.XKey[7]);
  end;
  n2:= n2 xor f(n1+Data.XKey[7]);
  n1:= n1 xor f(n2+Data.XKey[6]);
  n2:= n2 xor f(n1+Data.XKey[5]);
  n1:= n1 xor f(n2+Data.XKey[4]);
  n2:= n2 xor f(n1+Data.XKey[3]);
  n1:= n1 xor f(n2+Data.XKey[2]);
  n2:= n2 xor f(n1+Data.XKey[1]);
  n1:= n1 xor f(n2+Data.XKey[0]);
  Move(n2,OutData^,4);
  Move(n1,pointer(integer(OutData)+4)^,4);
end;

procedure GOSTDecryptECB;
var
  n1, n2: DWord;
  i: integer;
begin
  Move(InData^,n1,4);
  Move(pointer(integer(InData)+4)^,n2,4);
  n2:= n2 xor f(n1+Data.XKey[0]);
  n1:= n1 xor f(n2+Data.XKey[1]);
  n2:= n2 xor f(n1+Data.XKey[2]);
  n1:= n1 xor f(n2+Data.XKey[3]);
  n2:= n2 xor f(n1+Data.XKey[4]);
  n1:= n1 xor f(n2+Data.XKey[5]);
  n2:= n2 xor f(n1+Data.XKey[6]);
  n1:= n1 xor f(n2+Data.XKey[7]);
  for i:= 0 to 2 do
  begin
    n2:= n2 xor f(n1+Data.XKey[7]);
    n1:= n1 xor f(n2+Data.XKey[6]);
    n2:= n2 xor f(n1+Data.XKey[5]);
    n1:= n1 xor f(n2+Data.XKey[4]);
    n2:= n2 xor f(n1+Data.XKey[3]);
    n1:= n1 xor f(n2+Data.XKey[2]);
    n2:= n2 xor f(n1+Data.XKey[1]);
    n1:= n1 xor f(n2+Data.XKey[0]);
  end;
  Move(n2,OutData^,4);
  Move(n1,pointer(integer(OutData)+4)^,4);
end;

procedure GOSTEncryptCBC;
begin
  XorBlock(InData,@Data.LastBlock,OutData,8);
  GOSTEncryptECB(Data,OutData,OutData);
  Move(OutData^,Data.LastBlock,8);
end;

procedure GOSTDecryptCBC;
var
  TempBlock: array[0..7] of byte;
begin
  Move(InData^,TempBlock,8);
  GOSTDecryptECB(Data,InData,OutData);
  XorBlock(OutData,@Data.LastBlock,OutData,8);
  Move(TempBlock,Data.LastBlock,8);
end;

procedure GOSTEncryptCFB;
var
  i: integer;
  TempBlock: array[0..7] of byte;
begin
  for i:= 0 to Len-1 do
  begin
    GOSTEncryptECB(Data,@Data.LastBlock,@TempBlock);
    PByteArray(OutData)[i]:= PByteArray(InData)[i] xor TempBlock[0];
    Move(Data.LastBlock[1],Data.LastBlock[0],7);
    Data.LastBlock[7]:= PByteArray(OutData)[i];
  end;
end;

procedure GOSTDecryptCFB;
var
  i: integer;
  TempBlock: array[0..7] of byte;
  b: byte;
begin
  for i:= 0 to Len-1 do
  begin
    b:= PByteArray(InData)[i];
    GOSTEncryptECB(Data,@Data.LastBlock,@TempBlock);
    PByteArray(OutData)[i]:= PByteArray(InData)[i] xor TempBlock[0];
    Move(Data.LastBlock[1],Data.LastBlock[0],7);
    Data.LastBlock[7]:= b;
  end;
end;

procedure GOSTEncryptOFB;
begin
  GOSTEncryptECB(Data,@Data.LastBlock,@Data.LastBlock);
  XorBlock(@Data.LastBlock,InData,OutData,8);
end;

procedure GOSTDecryptOFB;
begin
  GOSTEncryptECB(Data,@Data.LastBlock,@Data.LastBlock);
  XorBlock(@Data.LastBlock,InData,OutData,8);
end;

procedure GOSTEncryptOFBC;
var
  i: integer;
  TempBlock: array[0..7] of byte;
begin
  for i:= 0 to Len-1 do
  begin
    GOSTEncryptECB(Data,@Data.LastBlock,@TempBlock);
    PByteArray(OutData)[i]:= PByteArray(InData)[i] xor TempBlock[0];
    IncBlock(@Data.LastBlock,8);
  end;
end;

procedure GOSTDecryptOFBC;
var
  i: integer;
  TempBlock: array[0..7] of byte;
begin
  for i:= 0 to Len-1 do
  begin
    GOSTEncryptECB(Data,@Data.LastBlock,@TempBlock);
    PByteArray(OutData)[i]:= PByteArray(InData)[i] xor TempBlock[0];
    IncBlock(@Data.LastBlock,8);
  end;
end;

procedure GOSTReset;
begin
  Move(Data.InitBlock,Data.LastBlock,8);
end;

end.
